/*
* funkcja tworzy pojedynczy shader
* gl - kontekst WebGL
* type - typ shadera: shader wierzcholkow (gl.VERTEX_SHADER), shader fragmentow (gl.FRAGMENT_SHADER)
* source - kod zrodlowy shadera (string)
* funkcja zwraca obiekt shadera (WebGLShader) lub null w razie niepowodzenia przy tworzeniu shadera
*/
function createShader(gl, type, source, id)
{
    let shader = gl.createShader( type );

    gl.shaderSource( shader, source );
    gl.compileShader( shader );

    if( !gl.getShaderParameter( shader, gl.COMPILE_STATUS ) )
    {
        let message = `Kompilacja shadera ${(type == gl.VERTEX_SHADER) ? "wierzchołków" : "fragmentów"} (${id}) nie powiodła się. Zawartość logu:`
            + `\n${gl.getShaderInfoLog( shader )}`;
            
        alert( message );

        gl.deleteShader( shader );

        return null;
    }

    return shader;
}

/*
* funkcja tworzy program cieniowania
* gl - kontekst WebGL
* vertexShaderId - identyfikator elementu strony, ktory zawiera kod zrodlowy shadera wierzcholkow
* fragmentShaderId - identyfikator elementy strony, ktory zawiera kod zrodlowy shadera fragmentow
* funkcja zwraca obiekt programu cieniowania (WebGLProgram) lub null w razie niepowodzenia przy tworzeniu programu cieniowania
*/
function createProgram(gl, vertexShaderId, fragmentShaderId)
{
    let vertexShader; // obiekt shadera wierzcholkow (WebGLShader)
    let fragmentShader; // obiekt shadera fragmentow (WebGLShader)

    let vertexSource = document.getElementById( vertexShaderId );
    let fragmentSource = document.getElementById( fragmentShaderId );

    if( !vertexSource )
    {
        alert( "Nie udało się załadować shadera wierzchołków " + vertexShaderId );

        return null;
    }
    else
    {
        if( !fragmentSource )
        {
            alert( "Nie udało się załadować shadera fragmentów " + fragmentShaderId );

            return null;
        }
    }

    vertexShader = createShader( gl, gl.VERTEX_SHADER, vertexSource.text, vertexShaderId );
    if( !vertexShader )
        return null;

    fragmentShader = createShader( gl, gl.FRAGMENT_SHADER, fragmentSource.text, fragmentShaderId );
    if( !fragmentShader )
        return null;

    let program = gl.createProgram();

    gl.attachShader( program, vertexShader );
    gl.attachShader( program, fragmentShader );

    gl.linkProgram( program );

    if( !gl.getProgramParameter( program, gl.LINK_STATUS ) )
    {
        let message = `Linkowanie programu cieniowania nie powiodło się. Zawartość logu:`
            + `\n${gl.getProgramInfoLog( program )}`;

        alert( message );

        gl.deleteShader( vertexShader );
        gl.deleteShader( fragmentShader );

        gl.deleteProgram( program );

        return null;
    }

    gl.deleteShader( vertexShader );
    gl.deleteShader( fragmentShader );

    return program;
}
