"use strict";

window.onload = init;

let gl; // kontekst WebGL

// dane o programie cieniowania z cieniowaniem Gouraud
let gouraudShading =
{
    "id" : null,
    "vPosition" : -1,
    "vNormal" : -1,
    "uModelViewMatrix" : null,
    "uNormalMatrix" : null,
    "uProjectionMatrix" : null,
    "uLightPosition" : null,
    "uLightAmbient" : null,
    "uLightDiffuse" : null,
    "uLightSpecular" : null,
    "uMaterialAmbient" : null,
    "uMaterialDiffuse" : null,
    "uMaterialSpecular" : null,
    "uMaterialShininess" : null
};

// dane o programie cieniowania z cieniowaniem Phonga
let phongShading =
{
    "id" : null,
    "vPosition" : -1,
    "vNormal" : -1,
    "uModelViewMatrix" : null,
    "uNormalMatrix" : null,
    "uProjectionMatrix" : null,
    "uLightPosition" : null,
    "uLightAmbient" : null,
    "uLightDiffuse" : null,
    "uLightSpecular" : null,
    "uMaterialAmbient" : null,
    "uMaterialDiffuse" : null,
    "uMaterialSpecular" : null,
    "uMaterialShininess" : null
};

let shaderProgram = gouraudShading; // aktualnie wybrany program cieniowania

// dane o zrodle swiatla
let light =
{
    "position" : glMatrix.vec4.fromValues( 0.0, 0.0, 5.0, 1.0 ), // w ukladzie swiata
    "ambient" : glMatrix.vec3.fromValues( 0.2, 0.2, 0.2 ),
    "diffuse" : glMatrix.vec3.fromValues( 1.0, 1.0, 1.0 ),
    "specular" : glMatrix.vec3.fromValues( 1.0, 1.0, 1.0 )
};

// dane o materiale
let material =
{
    "ambient" : glMatrix.vec3.fromValues( 1.0, 0.5, 0.0 ),
    "diffuse" : glMatrix.vec3.fromValues( 0.34615, 0.3143, 0.0903 ),
    "specular" : glMatrix.vec3.fromValues( 0.797357, 0.723991, 0.208006 ),
    "shininess" : 83.2
};

// dane o czajniczku
let teapot =
{
    "positionVBO" : null, // VBO zawierajace wspolrzedne
    "normalVBO" : null, // VBO zawierajace normale
    "indicesVBO" : null, // VBO zawierajace indeksy wierzcholkow scianek
    "noIndices" : 0, // liczba wierzcholkow do wyrenderowania
    "translationVector" : [0.0, 0.0, 0.0],
    "scaleFactors" : [1.0, 1.0, 1.0]
};

let modelViewMatrix = glMatrix.mat4.create(); // macierz model-widok
let projectionMatrix = glMatrix.mat4.create(); // macierz rzutowania
let rotationAngles = [0.0, 0.0, 0.0]; // katy obrotu wzgledem poszczegolnych osi

/*
* funkcja inicjuje WebGL oraz ustawia poczatkowe wartosci pewnych ustawien WebGL
*/
function init()
{
    initGL();
    setupShaders();
    setupBuffers();

    initUI();

    window.requestAnimFrame( render );
}

/*
* funkcja pobiera kontekst WebGL oraz inicjuje pewne ustawienia WebGL
*/
function initGL()
{
    let canvas = document.getElementById( "gl-canvas" );

    gl = WebGLUtils.setupWebGL( canvas );
    if( !gl )
    {
        alert( "WebGL jest niedostępny!" );
        throw new Error( "WebGL isn't available!" );
    }

    gl.viewport( 0, 0, canvas.width, canvas.height );
    gl.clearColor( 0.7, 0.7, 0.7, 1.0 );
    gl.enable( gl.DEPTH_TEST );

    glMatrix.mat4.perspective( projectionMatrix, Math.PI / 4, canvas.width / canvas.height, 0.1, 10.0 );
}

/*
* funkcja wczytuje shadery oraz pobiera lokalizacje atrybutow i zmiennych jednorodnych
*/
function setupShaders()
{
    gouraudShading.id = createProgram( gl, "gouraud_vertex-shader", "gouraud_fragment-shader" );

    if( !gouraudShading.id )
        throw new Error( "Shader program creation failed!" );

    gouraudShading.vPosition = gl.getAttribLocation( gouraudShading.id, "vPosition" );
    gouraudShading.vNormal = gl.getAttribLocation( gouraudShading.id, "vNormal" );

    gouraudShading.uModelViewMatrix = gl.getUniformLocation( gouraudShading.id, "uModelViewMatrix" );
    gouraudShading.uNormalMatrix = gl.getUniformLocation( gouraudShading.id, "uNormalMatrix" );
    gouraudShading.uProjectionMatrix = gl.getUniformLocation( gouraudShading.id, "uProjectionMatrix" );

    gouraudShading.uLightPosition = gl.getUniformLocation( gouraudShading.id, "uLight.position" );
    gouraudShading.uLightAmbient = gl.getUniformLocation( gouraudShading.id, "uLight.ambient" );
    gouraudShading.uLightDiffuse = gl.getUniformLocation( gouraudShading.id, "uLight.diffuse" );
    gouraudShading.uLightSpecular = gl.getUniformLocation( gouraudShading.id, "uLight.specular" );

    gouraudShading.uMaterialAmbient = gl.getUniformLocation( gouraudShading.id, "uMaterial.ambient" );
    gouraudShading.uMaterialDiffuse = gl.getUniformLocation( gouraudShading.id, "uMaterial.diffuse" );
    gouraudShading.uMaterialSpecular = gl.getUniformLocation( gouraudShading.id, "uMaterial.specular" );
    gouraudShading.uMaterialShininess = gl.getUniformLocation( gouraudShading.id, "uMaterial.shininess" );

    //---------------------------------------------------------

    phongShading.id = createProgram( gl, "phong_vertex-shader", "phong_fragment-shader" );

    if( !phongShading.id )
        throw new Error( "Shader program creation failed!" );

    phongShading.vPosition = gl.getAttribLocation( phongShading.id, "vPosition" );
    phongShading.vNormal = gl.getAttribLocation( phongShading.id, "vNormal" );

    phongShading.uModelViewMatrix = gl.getUniformLocation( phongShading.id, "uModelViewMatrix" );
    phongShading.uNormalMatrix = gl.getUniformLocation( phongShading.id, "uNormalMatrix" );
    phongShading.uProjectionMatrix = gl.getUniformLocation( phongShading.id, "uProjectionMatrix" );

    phongShading.uLightPosition = gl.getUniformLocation( phongShading.id, "uLight.position" );
    phongShading.uLightAmbient = gl.getUniformLocation( phongShading.id, "uLight.ambient" );
    phongShading.uLightDiffuse = gl.getUniformLocation( phongShading.id, "uLight.diffuse" );
    phongShading.uLightSpecular = gl.getUniformLocation( phongShading.id, "uLight.specular" );

    phongShading.uMaterialAmbient = gl.getUniformLocation( phongShading.id, "uMaterial.ambient" );
    phongShading.uMaterialDiffuse = gl.getUniformLocation( phongShading.id, "uMaterial.diffuse" );
    phongShading.uMaterialSpecular = gl.getUniformLocation( phongShading.id, "uMaterial.specular" );
    phongShading.uMaterialShininess = gl.getUniformLocation( phongShading.id, "uMaterial.shininess" );
}

/*
* funkcja tworzy obiekty VBO z czajniczkiem
*/
function setupBuffers()
{
    teapot.positionVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, teapot.positionVBO );
    gl.bufferData( gl.ARRAY_BUFFER, teapotPositions, gl.STATIC_DRAW );

    teapot.normalVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, teapot.normalVBO );
    gl.bufferData( gl.ARRAY_BUFFER, teapotNormals, gl.STATIC_DRAW );

    teapot.indicesVBO = gl.createBuffer();
    gl.bindBuffer( gl.ELEMENT_ARRAY_BUFFER, teapot.indicesVBO );
    gl.bufferData( gl.ELEMENT_ARRAY_BUFFER, teapotIndices, gl.STATIC_DRAW );

    teapot.noIndices = teapotIndices.length;

    calculateTranslationAndScale();
}

/*
* funkcja wyznacza wektor translacji i wspolczynnik skali tak, aby czajniczek
* znajdowal sie w poczatku ukladu wspolrzednych i miescil sie w szescianie o boku 2
*/
function calculateTranslationAndScale()
{
    let bbMin = [teapotPositions[0], teapotPositions[1], teapotPositions[2]];
    let bbMax = [teapotPositions[0], teapotPositions[1], teapotPositions[2]];
    let centroid = [0.0, 0.0, 0.0];
    let n = 0;

    for( let i = 0; i < teapotPositions.length; i += 3 )
    {
        centroid[0] += teapotPositions[i];
        centroid[1] += teapotPositions[i+1];
        centroid[2] += teapotPositions[i+2];
        ++n;

        for( let j = 0; j < 3; ++j )
        {
            if( teapotPositions[i + j] < bbMin[j] )
                bbMin[j] = teapotPositions[i + j];
                
            if( bbMax[j] < teapotPositions[i + j] )
                bbMax[j] = teapotPositions[i + j];
        }
    }

    centroid = centroid.map( v => v / n );
    teapot.translationVector = centroid.map( v => -v );

    let maxExtent = Math.max.apply( Math, [Math.abs(bbMax[0] - bbMin[0]), Math.abs(bbMax[1] - bbMin[1]), Math.abs(bbMax[2] - bbMin[2])] );
    let scale = 2.0 / maxExtent;
    teapot.scaleFactors = [scale, scale, scale];
}

/*
* funkcja oblicza iloczyn macierzy przez wektor, tzn. M v
* M - macierz
* v - wektor
*/
function multiplyMatrixByVector(M, v)
{
    let x = M[0] * v[0] + M[4] * v[1] + M[8] * v[2] + M[12] * v[3];
    let y = M[1] * v[0] + M[5] * v[1] + M[9] * v[2] + M[13] * v[3];
    let z = M[2] * v[0] + M[6] * v[1] + M[10] * v[2] + M[14] * v[3];
    let w = M[3] * v[0] + M[7] * v[1] + M[11] * v[2] + M[15] * v[3];

    return glMatrix.vec4.fromValues( x, y, z, w );
}

/*
* funkcja renderuje scene
*/
function render()
{
    gl.clear( gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT );

    gl.useProgram( shaderProgram.id );

    glMatrix.mat4.lookAt( modelViewMatrix, [0, 0, 3], [0, 0, 0], [0, 1, 0] );

    let lightPositionEye = multiplyMatrixByVector( modelViewMatrix, light.position ); // pozycja swiatla w ukladzie oka

    gl.uniform4fv( shaderProgram.uLightPosition, lightPositionEye );
    gl.uniform3fv( shaderProgram.uLightAmbient, light.ambient );
    gl.uniform3fv( shaderProgram.uLightDiffuse, light.diffuse );
    gl.uniform3fv( shaderProgram.uLightSpecular, light.specular );

    gl.uniform3fv( shaderProgram.uMaterialAmbient, material.ambient );
    gl.uniform3fv( shaderProgram.uMaterialDiffuse, material.diffuse );
    gl.uniform3fv( shaderProgram.uMaterialSpecular, material.specular );
    gl.uniform1f( shaderProgram.uMaterialShininess, material.shininess );

    gl.uniformMatrix4fv( shaderProgram.uProjectionMatrix, false, projectionMatrix );

    glMatrix.mat4.scale( modelViewMatrix, modelViewMatrix, teapot.scaleFactors );
    glMatrix.mat4.rotateZ( modelViewMatrix, modelViewMatrix, rotationAngles[2] * Math.PI / 180.0 );
    glMatrix.mat4.rotateY( modelViewMatrix, modelViewMatrix, rotationAngles[1] * Math.PI / 180.0 );
    glMatrix.mat4.rotateX( modelViewMatrix, modelViewMatrix, rotationAngles[0] * Math.PI / 180.0 );
    glMatrix.mat4.translate( modelViewMatrix, modelViewMatrix, teapot.translationVector );

    let normalMatrix = glMatrix.mat3.create();
    glMatrix.mat3.normalFromMat4( normalMatrix, modelViewMatrix );

    gl.uniformMatrix4fv( shaderProgram.uModelViewMatrix, false, modelViewMatrix );
    gl.uniformMatrix3fv( shaderProgram.uNormalMatrix, false, normalMatrix );

    renderTeapot();

    window.requestAnimFrame( render );
}

/*
* funkcja rysujaca czajniczek
*/
function renderTeapot()
{
    gl.enableVertexAttribArray( shaderProgram.vPosition );
    gl.enableVertexAttribArray( shaderProgram.vNormal );

    gl.bindBuffer( gl.ARRAY_BUFFER, teapot.positionVBO );
    gl.vertexAttribPointer( shaderProgram.vPosition, 3, gl.FLOAT, false, 0, 0 );
    gl.bindBuffer( gl.ARRAY_BUFFER, teapot.normalVBO );
    gl.vertexAttribPointer( shaderProgram.vNormal, 3, gl.FLOAT, false, 0, 0 );

    gl.bindBuffer( gl.ELEMENT_ARRAY_BUFFER, teapot.indicesVBO );
    gl.drawElements( gl.TRIANGLES, teapot.noIndices, gl.UNSIGNED_SHORT, 0 );
}