"use strict";

window.onload = init;

let gl; // kontekst WebGL

// dane o programie cieniowania z cieniowaniem Phonga
let phongShading =
{
    "id" : null,
    "vPosition" : -1,
    "vNormal" : -1,
    "uModelViewMatrix" : null,
    "uNormalMatrix" : null,
    "uProjectionMatrix" : null,
    "uLightPosition" : null,
    "uLightAmbient" : null,
    "uLightDiffuse" : null,
    "uLightSpecular" : null,
    "uMaterialAmbient" : null,
    "uMaterialDiffuse" : null,
    "uMaterialSpecular" : null,
    "uMaterialShininess" : null
};

// dane o zrodle swiatla
let light =
{
    "position" : glMatrix.vec4.fromValues( 0.0, 0.0, 5.0, 1.0 ), // w ukladzie swiata
    "ambient" : glMatrix.vec3.fromValues( 0.2, 0.2, 0.2 ),
    "diffuse" : glMatrix.vec3.fromValues( 1.0, 1.0, 1.0 ),
    "specular" : glMatrix.vec3.fromValues( 1.0, 1.0, 1.0 )
};

// dane o materiale
let material =
{
    "ambient" : glMatrix.vec3.fromValues( 1.0, 0.5, 0.0 ),
    "diffuse" : glMatrix.vec3.fromValues( 0.34615, 0.3143, 0.0903 ),
    "specular" : glMatrix.vec3.fromValues( 0.797357, 0.723991, 0.208006 ),
    "shininess" : 83.2
};

// dane o szescianie z normalami ustawionymi per face
let cubePerFace =
{
    "positionVBO" : null,
    "normalVBO" : null,
    "noVertices" : 0,
    "render" : renderCubePerFace
};

// dane o szescianie z normalami ustawionymi per vertex
let cubePerVertex =
{
    "positionVBO" : null,
    "normalVBO" : null,
    "indicesVBO" : null,
    "noIndices" : 0,
    "render" : renderCubePerVertex
};

let modelViewMatrix = glMatrix.mat4.create(); // macierz model-widok
let projectionMatrix = glMatrix.mat4.create(); // macierz rzutowania
let rotationAngles = [0.0, 0.0, 0.0]; // katy obrotu wzgledem poszczegolnych osi

let model = cubePerFace; // aktualnie wybrany model

/*
* funkcja inicjuje WebGL oraz ustawia poczatkowe wartosci pewnych ustawien WebGL
*/
function init()
{
    initGL();
    setupShaders();
    setupBuffers();

    initUI();

    window.requestAnimFrame( render );
}

/*
* funkcja pobiera kontekst WebGL oraz inicjuje pewne ustawienia WebGL
*/
function initGL()
{
    let canvas = document.getElementById( "gl-canvas" );

    gl = WebGLUtils.setupWebGL( canvas );
    if( !gl )
    {
        alert( "WebGL jest niedostępny!" );
        throw new Error( "WebGL isn't available!" );
    }

    gl.viewport( 0, 0, canvas.width, canvas.height );
    gl.clearColor( 0.7, 0.7, 0.7, 1.0 );
    gl.enable( gl.DEPTH_TEST );

    glMatrix.mat4.perspective( projectionMatrix, Math.PI / 4, canvas.width / canvas.height, 0.1, 10.0 );
}

/*
* funkcja wczytuje shadery oraz pobiera lokalizacje atrybutow i zmiennych jednorodnych
*/
function setupShaders()
{
    phongShading.id = createProgram( gl, "phong_vertex-shader", "phong_fragment-shader" );

    if( !phongShading.id )
        throw new Error( "Shader program creation failed!" );

    phongShading.vPosition = gl.getAttribLocation( phongShading.id, "vPosition" );
    phongShading.vNormal = gl.getAttribLocation( phongShading.id, "vNormal" );

    phongShading.uModelViewMatrix = gl.getUniformLocation( phongShading.id, "uModelViewMatrix" );
    phongShading.uNormalMatrix = gl.getUniformLocation( phongShading.id, "uNormalMatrix" );
    phongShading.uProjectionMatrix = gl.getUniformLocation( phongShading.id, "uProjectionMatrix" );

    phongShading.uLightPosition = gl.getUniformLocation( phongShading.id, "uLight.position" );
    phongShading.uLightAmbient = gl.getUniformLocation( phongShading.id, "uLight.ambient" );
    phongShading.uLightDiffuse = gl.getUniformLocation( phongShading.id, "uLight.diffuse" );
    phongShading.uLightSpecular = gl.getUniformLocation( phongShading.id, "uLight.specular" );

    phongShading.uMaterialAmbient = gl.getUniformLocation( phongShading.id, "uMaterial.ambient" );
    phongShading.uMaterialDiffuse = gl.getUniformLocation( phongShading.id, "uMaterial.diffuse" );
    phongShading.uMaterialSpecular = gl.getUniformLocation( phongShading.id, "uMaterial.specular" );
    phongShading.uMaterialShininess = gl.getUniformLocation( phongShading.id, "uMaterial.shininess" );
}

/*
* funkcja tworzy obiekty VBO z modelami
*/
function setupBuffers()
{
    let cubePF = generateCubePerFace();

    cubePerFace.positionVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, cubePerFace.positionVBO );
    gl.bufferData( gl.ARRAY_BUFFER, new Float32Array( cubePF.vertices ), gl.STATIC_DRAW );
    gl.enableVertexAttribArray( phongShading.vPosition );

    cubePerFace.normalVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, cubePerFace.normalVBO );
    gl.bufferData( gl.ARRAY_BUFFER, new Float32Array( cubePF.normals ), gl.STATIC_DRAW );
    gl.enableVertexAttribArray( phongShading.vNormal );

    cubePerFace.noVertices = cubePF.vertices.length / 3.0;

    //-----------------------------------------------------

    let cubePV = generateCubePerVertex();

    cubePerVertex.positionVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, cubePerVertex.positionVBO );
    gl.bufferData( gl.ARRAY_BUFFER, new Float32Array( cubePV.vertices ), gl.STATIC_DRAW );
    gl.enableVertexAttribArray( phongShading.vPosition );

    cubePerVertex.normalVBO = gl.createBuffer();
    gl.bindBuffer( gl.ARRAY_BUFFER, cubePerVertex.normalVBO );
    gl.bufferData( gl.ARRAY_BUFFER, new Float32Array( cubePV.normals ), gl.STATIC_DRAW );
    gl.enableVertexAttribArray( phongShading.vNormal );

    cubePerVertex.indicesVBO = gl.createBuffer();
    gl.bindBuffer( gl.ELEMENT_ARRAY_BUFFER, cubePerVertex.indicesVBO );
    gl.bufferData( gl.ELEMENT_ARRAY_BUFFER, new Uint16Array( cubePV.indices ), gl.STATIC_DRAW );

    cubePerVertex.noIndices = cubePV.indices.length;
}

/*
* funkcja generuje dane szescianu z normalami ustawionymi per face
*/
function generateCubePerFace()
{
    let vertices =
    [
        // przod
        -1, -1, 1, 1, 1, 1, -1, 1, 1, -1, -1, 1, 1, -1, 1, 1, 1, 1,
        // prawy bok
        1, -1, 1, 1, 1, -1, 1, 1, 1, 1, -1, 1, 1, -1, -1, 1, 1, -1,
        // tyl
        1, -1, -1, -1, 1, -1, 1, 1, -1, 1, -1, -1, -1, -1, -1, -1, 1, -1,
        // lewy bok
        -1, -1, -1, -1, 1, 1, -1, 1, -1, -1, -1, -1, -1, -1, 1, -1, 1, 1,
        // gora
        -1, 1, 1, 1, 1, -1, -1, 1, -1, -1, 1, 1, 1, 1, 1, 1, 1, -1,
        // dol
        -1, -1, -1, 1, -1, 1, -1, -1, 1, -1, -1, -1, 1, -1, -1, 1, -1, 1
    ];

    let normals =
    [
        // przod
        0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1,
        // prawy bok
        1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0,
        // tyl
        0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1,
        // lewy bok
        -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0,
        // gora
        0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0,
        // dol
        0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0, 0, -1, 0
    ];

    let result =
    {
        "vertices" : vertices,
        "normals" : normals
    };

    return result;
}

/*
* funkcja generuje dane szescianu z normalami ustawionymi per vertex
*/
function generateCubePerVertex()
{
    let vertices =
    [
        -1, -1, 1,
        1, -1, 1,
        1, 1, 1,
        -1, 1, 1,
        1, -1, -1,
        -1, -1, -1,
        -1, 1, -1,
        1, 1, -1
    ];

    let s = 1.0 / Math.sqrt( 3.0 );
    let normals =
    [
        -s, -s, s,
        s, -s, s,
        s, s, s,
        -s, s, s,
        s, -s, -s,
        -s, -s, -s,
        -s, s, -s,
        s, s, -s
    ];

    let indices =
    [
        0, 1, 2,
        0, 2, 3,
        1, 4, 7,
        1, 7, 2,
        4, 5, 6,
        4, 6, 7,
        5, 0, 3,
        5, 3, 6,
        3, 2, 7,
        3, 7, 6,
        5, 4, 1,
        5, 1, 0
    ];

    let result =
    {
        "vertices" : vertices,
        "normals" : normals,
        "indices" : indices
    };

    return result;
}

/*
* funkcja oblicza iloczyn macierzy przez wektor, tzn. M v
* M - macierz
* v - wektor
*/
function multiplyMatrixByVector(M, v)
{
    let x = M[0] * v[0] + M[4] * v[1] + M[8] * v[2] + M[12] * v[3];
    let y = M[1] * v[0] + M[5] * v[1] + M[9] * v[2] + M[13] * v[3];
    let z = M[2] * v[0] + M[6] * v[1] + M[10] * v[2] + M[14] * v[3];
    let w = M[3] * v[0] + M[7] * v[1] + M[11] * v[2] + M[15] * v[3];

    return glMatrix.vec4.fromValues( x, y, z, w );
}

/*
* funkcja renderuje scene
*/
function render()
{
    gl.clear( gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT );

    gl.useProgram( phongShading.id );

    gl.enableVertexAttribArray( phongShading.vPosition );
    gl.enableVertexAttribArray( phongShading.vNormal );

    glMatrix.mat4.lookAt( modelViewMatrix, [0, 0, 5], [0, 0, 0], [0, 1, 0] );

    let lightPositionEye = multiplyMatrixByVector( modelViewMatrix, light.position ); // pozycja swiatla w ukladzie oka

    gl.uniform4fv( phongShading.uLightPosition, lightPositionEye );
    gl.uniform3fv( phongShading.uLightAmbient, light.ambient );
    gl.uniform3fv( phongShading.uLightDiffuse, light.diffuse );
    gl.uniform3fv( phongShading.uLightSpecular, light.specular );

    gl.uniform3fv( phongShading.uMaterialAmbient, material.ambient );
    gl.uniform3fv( phongShading.uMaterialDiffuse, material.diffuse );
    gl.uniform3fv( phongShading.uMaterialSpecular, material.specular );
    gl.uniform1f( phongShading.uMaterialShininess, material.shininess );

    gl.uniformMatrix4fv( phongShading.uProjectionMatrix, false, projectionMatrix );

    glMatrix.mat4.rotateZ( modelViewMatrix, modelViewMatrix, rotationAngles[2] * Math.PI / 180.0 );
    glMatrix.mat4.rotateY( modelViewMatrix, modelViewMatrix, rotationAngles[1] * Math.PI / 180.0 );
    glMatrix.mat4.rotateX( modelViewMatrix, modelViewMatrix, rotationAngles[0] * Math.PI / 180.0 );

    let normalMatrix = glMatrix.mat3.create();
    glMatrix.mat3.normalFromMat4( normalMatrix, modelViewMatrix );

    gl.uniformMatrix4fv( phongShading.uModelViewMatrix, false, modelViewMatrix );
    gl.uniformMatrix3fv( phongShading.uNormalMatrix, false, normalMatrix );

    model.render();

    window.requestAnimFrame( render );
}

/*
* funkcja rysujaca szescian z normalami ustawionymi per face
*/
function renderCubePerFace()
{
    gl.bindBuffer( gl.ARRAY_BUFFER, this.positionVBO );
    gl.vertexAttribPointer( phongShading.vPosition, 3, gl.FLOAT, false, 0, 0 );

    gl.bindBuffer( gl.ARRAY_BUFFER, this.normalVBO );
    gl.vertexAttribPointer( phongShading.vNormal, 3, gl.FLOAT, false, 0, 0 );

    gl.drawArrays( gl.TRIANGLES, 0, this.noVertices );
}

/*
* funkcja rysujaca szescian z normalami ustawionymi per vertex
*/
function renderCubePerVertex()
{
    gl.bindBuffer( gl.ARRAY_BUFFER, this.positionVBO );
    gl.vertexAttribPointer( phongShading.vPosition, 3, gl.FLOAT, false, 0, 0 );

    gl.bindBuffer( gl.ARRAY_BUFFER, this.normalVBO );
    gl.vertexAttribPointer( phongShading.vNormal, 3, gl.FLOAT, false, 0, 0 );

    gl.bindBuffer( gl.ELEMENT_ARRAY_BUFFER, this.indicesVBO );

    gl.drawElements( gl.TRIANGLES, this.noIndices, gl.UNSIGNED_SHORT, 0 );
}